import inspect
import json
import pydoc
from pathlib import Path
from pydoc import render_doc
from types import ModuleType
from typing import Any

import pandas as pd

INVOKING_MODULE_FRAME_IDX = 2


def get_documentation(func_or_cls):
    """Get the output printed by the `help` function as a string"""
    return '\n'.join(render_doc(func_or_cls).splitlines()[2:])


def get_docstring(functionality: object):
    methods_docstring = '\n\n'.join(
        get_documentation(method)
        for method_name, method in functionality.__class__.__dict__.items()
        if not method_name.startswith('_') and callable(method)
    )
    functionality_cls_docstring = functionality.__doc__ if functionality.__doc__ else ""
    return functionality_cls_docstring + '\n' + f"Functions defined here:" + '\n\n' + methods_docstring


class FunctionalModule(ModuleType):
    """
    Hack - Python versions < 3.7 modules cannot implement __getattr__ and __dir__
    """

    def __init__(self, functionality: object, module: ModuleType = None):
        self._module = module if module is not None else self._get_invoking_module()
        self._functionality = functionality
        super().__init__(self._module.__name__, get_docstring(self._functionality))

    def _get_invoking_module(self) -> ModuleType:
        """
        Current call stack illustration:
        ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
        Fr.#0  ║  self._get_invoking_module()          ║ (Current)
        ‾‾‾‾‾  ╠═══════════════════════════════════════╣
        Fr.#1  ║  self.__init__()                      ║
        ‾‾‾‾‾  ╠═══════════════════════════════════════╣
        Fr.#2  ║  module we're actually interested in  ║ (INVOKING_MODULE_FRAME_IDX = 2)
        ‾‾‾‾‾  ╠═══════════════════════════════════════╣
         ...   ║  ...                                  ║
               ╚═══════════════════════════════════════╝
        """
        invoking_frame = inspect.stack()[INVOKING_MODULE_FRAME_IDX]
        invoking_module = inspect.getmodule(invoking_frame[0])
        return invoking_module

    def __getattr__(self, name):
        try:
            return getattr(self._functionality, name)
        except AttributeError:
            return getattr(self._module, name)


def get_resource_path(*path_components: str) -> Path:
    invoking_frame = inspect.stack()[1]
    invoking_module = inspect.getmodule(invoking_frame[0])
    return Path(invoking_module.__file__).parent.joinpath(*path_components)


def load_resource(*path_components: str) -> Any:
    invoking_frame = inspect.stack()[1]
    invoking_module = inspect.getmodule(invoking_frame[0])
    pkg_resource_file_path = Path(invoking_module.__file__).parent.joinpath(*path_components)
    if pkg_resource_file_path.suffix == ".json":
        return json.loads(pkg_resource_file_path.read_text())
    elif pkg_resource_file_path.suffix == ".csv":
        return pd.read_csv(pkg_resource_file_path)



from http import HTTPStatus

from datagen.dev.logging import get_logger

logger = get_logger(__name__)


class ClientException(Exception):
    message = ""

    def __init__(self, response_msg: str, generation_id: str = None):
        self.msg = self.message.format(response_msg=response_msg, generation_id=generation_id)


class InvalidRequest(ClientException):
    message = "Invalid request error. Failed with error: {response_msg}"


class AuthenticationError(ClientException):
    message = "Authentication error. Authentication token is invalid or expired." " Failed with error: {response_msg}"


class GenerationIdNotFoundError(ClientException):
    message = "Generation ID: {generation_id} not found error. Failed with error: {response_msg}"


class HttpStatusHandler:
    status_exception_map = {
        HTTPStatus.FORBIDDEN: AuthenticationError,
        HTTPStatus.NOT_FOUND: GenerationIdNotFoundError,
        HTTPStatus.INTERNAL_SERVER_ERROR: InvalidRequest,
    }

    @classmethod
    def handle(cls, status_code: int, message: str = None) -> None:
        exception_class = cls.status_exception_map.get(status_code)
        if exception_class:
            logger.error(message)
            raise exception_class(message)
        else:
            raise ClientException("Something when wrong. Please try again later.")



import json
from pathlib import Path
from typing import List, Optional, Union

from datagen.api.assets import Background, Camera, DataRequest, Glasses, Human, HumanDatapoint, Light, Mask

# from datagen.api.client.impl import DataGenerationClient
# from datagen.api.client.schemas import DataResponse, DataResponseStatus
from datagen.api.requests.datapoint.builder import HumanDatapointBuilder
from datagen.api.requests.director import DataRequestDirector
from datagen.dev.logging import get_logger

logger = get_logger(__name__)

DEFAULT_DATAPOINT_REQUESTS_JSON_NAME = "datagen_data_request.json"


class DatagenAPI:
    def __init__(
        self,
        # client: DataGenerationClient,
    ):
        self._request_director = DataRequestDirector()
        # self._client = client

    def create_datapoint(
        self,
        human: Human,
        camera: Camera,
        glasses: Optional[Glasses] = None,
        mask: Optional[Mask] = None,
        background: Optional[Background] = None,
        lights: Optional[List[Light]] = None,
    ) -> HumanDatapoint:
        self._request_director.builder = HumanDatapointBuilder(human, camera, glasses, mask, background, lights)
        return self._request_director.build_datapoint()

    # def generate(self, request: DataRequest) -> DataResponse:
    #     return self._client.generate(request)
    #
    # def stop(self, generation_id: str) -> None:
    #     return self._client.stop(generation_id=generation_id)
    #
    # def get_status(self, generation_id: str) -> DataResponseStatus:
    #     return self._client.get_status(generation_id=generation_id)
    #
    # def get_download_urls(self, generation_id: str) -> List[str]:
    #     return self._client.get_download_urls(generation_id=generation_id)
    #
    # def download(self, urls: List[str], dest: str) -> None:
    #     self._client.download(urls=urls, dest=Path(dest))

    def load(self, path: Union[Path, str]) -> DataRequest:
        if not isinstance(path, Path):
            path = self._get_request_json_path(path)
        path = self._get_request_json_path(path)
        return DataRequest.parse_file(path)

    def dump(self, request: DataRequest, path: Union[Path, str] = None) -> None:
        if not isinstance(path, Path):
            path = self._get_request_json_path(path)
        path.write_text(json.dumps(request.dict(), indent=3, sort_keys=True))
        logger.info(
            f"Request was successfully dumped to path '{path.absolute()}' ({len(request)} datapoints total).",
        )

    def _get_request_json_path(self, path: Optional[str]) -> Path:
        path_ = Path(path) if path else Path.cwd().joinpath(DEFAULT_DATAPOINT_REQUESTS_JSON_NAME)
        if not path_.parent.exists():
            raise FileNotFoundError(f"{path_.parent} folder does not exist, cannot dump requests")
        return path_



from dependency_injector import containers, providers

from datagen.modalities.textual.base.containers import (
    BaseModalitiesContainer,
    modality_dataclass_factory,
    modality_factory,
)


class ActorsModalityContainer(containers.DeclarativeContainer):

    from .actor_metadata import Actors

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=Actors)})


class KeypointsModalityContainer(containers.DeclarativeContainer):

    from .keypoints import Keypoints

    create = providers.FactoryAggregate(
        {1: modality_dataclass_factory(clazz=Keypoints), 2: modality_dataclass_factory(clazz=Keypoints)}
    )


class CenterOfGeometryModalityContainer(containers.DeclarativeContainer):

    from .center_of_geometry import CenterOfGeometry

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=CenterOfGeometry)})


class HICModalitiesContainer(BaseModalitiesContainer):

    actors = providers.Callable(modality_factory, modality_container=providers.Container(ActorsModalityContainer))

    keypoints = providers.Callable(modality_factory, modality_container=providers.Container(KeypointsModalityContainer))

    center_of_geometry = providers.Callable(
        modality_factory, modality_container=providers.Container(CenterOfGeometryModalityContainer)
    )



import os
import numpy as np

from datagen.imaging.base import ImagingLibrary

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"

import cv2


class OpenCVImagingLibrary(ImagingLibrary):
    def _read_png(self, image_file_path: str, keep_alpha: bool, convert_to_uint8: bool) -> np.ndarray:
        img = cv2.imread(image_file_path, cv2.IMREAD_UNCHANGED)
        if keep_alpha:
            img[..., :3] = cv2.cvtColor(img[..., :3], cv2.COLOR_BGR2RGB)
        else:
            img = cv2.cvtColor(img[..., :3], cv2.COLOR_BGR2RGB)
        if convert_to_uint8:
            img = self._convert_to_uint8(img)
        return img

    @staticmethod
    def _convert_to_uint8(img: np.ndarray) -> np.ndarray:
        if img.dtype == np.uint16:
            img = ((img / (2 ** 16 - 1)) * 255).astype(np.uint8)
        elif img.dtype == np.uint8:
            return img
        else:
            raise ValueError("Cannot convert image to uint8 format")
        return img

    def _read_exr(self, image_file_path: str) -> np.ndarray:
        img = cv2.imread(image_file_path, cv2.IMREAD_UNCHANGED)
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)



from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List, Tuple

from datagen.components.scene import Scene

# Older datasets scenes are named "environment_XXXXX".
SCENE_FOLDER_NAME_PATTERNS = ["scene_*", "environment_*", "datapoint_*"]


class CorruptedSourceError(ValueError):
    ...


@dataclass
class DataSource:
    path: Path

    @property
    def environment(self) -> str:
        return "hic" if self._is_hic_datasource() else "identities"

    def init_scenes(self) -> List[Scene]:
        scenes = [Scene(path=path_) for path_ in sorted(self.path.iterdir()) if self._is_path_to_scene(path_)]
        if len(scenes) == 0:
            raise CorruptedSourceError(f"Corrupted source: {str(self.path)}")
        return scenes

    def _is_hic_datasource(self) -> bool:
        scenes_paths = filter(lambda p: self._is_path_to_scene(p), self.path.iterdir())
        return all(Scene.is_path_to_hic_scene(scene_path) for scene_path in scenes_paths)

    @staticmethod
    def _is_path_to_scene(path: Path) -> bool:
        return any(path.match(pattern) for pattern in SCENE_FOLDER_NAME_PATTERNS) and path.is_dir()


@dataclass
class SourcesRepository:
    sources_paths: Tuple[str]

    def get_all(self) -> Iterator[DataSource]:
        for path in self.sources_paths:
            yield DataSource(path=Path(path))



import logging
import sys
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

import appdirs

FORMATTER = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    dir_path = appdirs.user_log_dir(appname=".datagen")
    Path(dir_path).mkdir(parents=True, exist_ok=True)
    log_file_path = Path(dir_path + "datagen.log")
    log_file_path.touch(exist_ok=True)
    file_handler = TimedRotatingFileHandler(filename=log_file_path, when="midnight")
    file_handler.setFormatter(fmt=FORMATTER)
    return file_handler


def get_logger(logger_name: str):
    logger = logging.getLogger(name=logger_name)
    logger.setLevel(level=logging.INFO)
    logger.addHandler(hdlr=get_console_handler())
    logger.addHandler(hdlr=get_file_handler())
    logger.propagate = False
    return logger



from datagen.components.datasource import DataSource
from datagen.components.datasource import SourcesRepository
from datagen.components.camera import Camera
from datagen.components.scene import Scene
from datagen.components.dataset import Dataset



import sys

from .containers import DatagenContainer
from .components.dataset import DatasetConfig
from datagen.dev import __version__, FunctionalModule

datagen = DatagenContainer().datagen()

sys.modules[__name__] = FunctionalModule(datagen)



import enum

from pydantic import BaseModel, Field


class DataResponse(BaseModel):
    generation_name: str = Field(
        description="Generation name provided by Datagen platform. In a format " "of 'Field - <date>."
    )
    generation_id: str = Field(
        description="Unique UUID provided by Datagen platform. With the generation id you'll "
        "be able to query generation_status, stop generation and get download urls"
    )
    dgu_hour: float = Field(description="The DGU Hour cost of the requested generation.")
    renders: int = Field(description="The number of datapoints received in this request.")
    scenes: int = Field(description="The number of scenes received in this request.")


class EGenerationStatus(enum.Enum):
    START = "START"
    IN_PROGRESS = "IN-PROGRESS"
    PENDING = "IN-PENDING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class DataResponseStatus(BaseModel):
    estimation_time_ms: int = Field(description="Estimated time in ms to data generation completion.", ge=0)
    status: EGenerationStatus = Field(description="Current status of the generation process.")
    percentage: int = Field(description="The percentage of data generation completion.", ge=0, le=100)



from dataclasses import dataclass, field
from pathlib import Path

from datagen import modalities
from datagen.modalities.containers import DatapointModalitiesContainer
from datagen.modalities.textual.base.environments import Environment


@dataclass
class DataPoint:
    visible_spectrum_image_name: str
    frame_num: int
    camera: str
    scene_path: Path
    modalities_container: DatapointModalitiesContainer = field(repr=False)

    @property
    def frame_path(self) -> Path:
        return self.scene_path.joinpath("frames", str(self.frame_num).zfill(3))

    @property
    def camera_path(self) -> Path:
        if self.frame_path.exists():
            return self.frame_path.joinpath(self.camera)
        else:
            return self.scene_path.joinpath(self.camera)

    @modalities.visual_modality
    def visible_spectrum(self) -> modalities.VisualModality:
        return modalities.VisualModality(
            self.visible_spectrum_image_name,
            keep_alpha=self.environment.transparent_background,
            convert_to_uint8=True,
        )

    @modalities.visual_modality
    def semantic_segmentation(self) -> modalities.VisualModality:
        return modalities.VisualModality("semantic_segmentation.png")

    @modalities.textual_modality
    def semantic_segmentation_metadata(self) -> modalities.TextualModality:
        return modalities.TextualModality(factory_name="segmentation", file_name="semantic_segmentation_metadata.json")

    @modalities.visual_modality
    def infrared_spectrum(self) -> modalities.VisualModality:
        return modalities.VisualModality("infrared_spectrum.png", convert_to_uint8=True)

    @modalities.visual_modality
    def depth(self) -> modalities.VisualModality:
        return modalities.VisualModality("depth.exr")

    @modalities.visual_modality
    def normal_maps(self) -> modalities.VisualModality:
        return modalities.VisualModality("normal_maps.exr")

    @modalities.textual_modality
    def camera_metadata(self) -> modalities.TextualModality:
        return modalities.TextualModality(factory_name="camera_metadata", file_name="camera_metadata.json")

    @modalities.textual_modality
    def _environments(self) -> modalities.TextualModality:
        return modalities.TextualModality(factory_name="environments", file_name="environment.json")

    @property
    def environment(self) -> Environment:
        return self._environments[self.visible_spectrum_image_name]

    @modalities.textual_modality
    def lights_metadata(self) -> modalities.TextualModality:
        return modalities.TextualModality(factory_name="lights_metadata", file_name="lights_metadata.json")



from datagen.modalities.textual.base.containers import BaseModalitiesContainer
from datagen.modalities.textual.identities.containers import IdentitiesModalitiesContainer
from datagen.modalities.textual.hic.containers import HICModalitiesContainer



import concurrent.futures
import json
import urllib.parse
from dataclasses import dataclass
from http import HTTPStatus
from pathlib import Path
from typing import List

import requests

from datagen.api.assets import DataRequest, HumanDatapoint
from datagen.api.client.exceptions import HttpStatusHandler
from datagen.api.client.schemas import DataResponse, DataResponseStatus
from datagen.config import settings
from datagen.dev.logging import get_logger

logger = get_logger(__name__)


@dataclass
class AuthenticationConfig:
    token: str
    user_id: str
    org_id: str


class DataGenerationClient:
    def __init__(self, config: AuthenticationConfig):
        self._url = settings["url"]
        self._headers = dict(authorizationToken=config.token)
        self._user_id = config.user_id
        self._org_id = config.org_id

    def generate(self, request: DataRequest) -> DataResponse:
        """
        Send a data generation request to Datagen's services.
        This request will trigger a data generation process that will eventually will result in an url
        link to a s3 bucket that contains the generated data.
        :param request: Data generation request that consists of all the data point requests.
        :return: Unique ID that represents the generation id of this data generation request.
        :raise InvalidRequest: In case of an invalid request.
        """
        if any(map(lambda x: not isinstance(x, HumanDatapoint), request.datapoints)):
            raise TypeError("Invalid datapoint type was provided.")
        else:
            return self._generate_human_identities_data(request=request)

    def get_status(self, generation_id: str) -> DataResponseStatus:
        """
        Data generation status consists of the following:
        1. Time estimation in ms
        2. Current generation status e.g. IN_PROGRESS, COMPLETED, FAILED, etc.
        3. Percentage of the data generation process.
        :param generation_id: Generation ID that was provided by generate_data call.
        :return: Data generation status.
        :raise: GenerationIdNotFoundError in case of an invalid generation id
        """
        response = requests.get(url=self._url + f"/generations/status/{generation_id}", headers=self._headers)
        if response.status_code != HTTPStatus.OK:
            message = f"Failed to get generation status of {generation_id}."
            HttpStatusHandler.handle(status_code=response.status_code, message=message)

        response_data = json.loads(response.content)
        return DataResponseStatus(
            estimation_time_ms=response_data["estimatedTime"],
            status=response_data["status"],
            percentage=response_data["percentage"],
        )

    def get_download_urls(self, generation_id: str) -> List[str]:
        """
        Once data generation of a generation_id is done, a download link will be provided.
        :param generation_id: Generation ID that was provided by generate_data call.
        :return: Download links to generated data.
        :raise: GenerationIdNotFoundError in case of an invalid generation id.
        :raise: InvalidRequest in case of an invalid request.
        """
        response = requests.get(url=self._url + f"delivery/{generation_id}", headers=self._headers)
        if response.status_code == HTTPStatus.ACCEPTED:
            logger.info(
                f"Data generation is in progress. Download links for generation ID {generation_id} will soon be ready."
            )
            return []

        elif response.status_code != HTTPStatus.OK:
            message = f"Failed to get s3 url for get download urls for {generation_id}."
            HttpStatusHandler.handle(status_code=response.status_code, message=message)

        logger.info(f"Successfully received download links for generation id: {generation_id}.")
        return json.loads(response.content)

    def stop(self, generation_id: str) -> None:
        """
        Stops data generation request.
        A table of all generation requests and their progress can be found here:
        https://app.prod.datagen.tech/dataset.

        :param generation_id: Generation ID that was provided by generate_data call.
        :raise: InvalidRequest in case of an invalid request.
        """
        response = requests.post(url=self._url + f"/generations/stop/{generation_id}", headers=self._headers)
        if response.status_code != HTTPStatus.OK:
            message = f"Failed to stop generation: {generation_id}."
            HttpStatusHandler.handle(status_code=response.status_code, message=message)

        logger.info(f"Successfully stopped data generation request: {generation_id}.")

    @classmethod
    def download(cls, urls: List[str], dest: Path) -> None:
        """
        Download the datasets provided by the get_download_urls method.
        Links are available for a limited time. If links are expired use get_download_urls to get valid urls.
        :param urls: Download urls received from get_download_urls method.
        :param dest: Directory to download the datasets to. If
        """
        if not dest.exists() and dest.is_dir():
            logger.error(f"Failed to download files to {str(dest)}. Directory does not exists.")
            raise FileExistsError(f"Directory {str(dest)} does not exists.")

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = [executor.submit(DataGenerationClient._download_file, url) for url in urls]

            for future in concurrent.futures.as_completed(results):
                url = urls[results.index(future)]
                try:
                    data = future.result()
                except Exception as e:
                    logger.error(f"{url} generated an exception: {e}")

                else:
                    filename = DataGenerationClient._extract_filename(url=url)
                    with open(f"{str(dest)}/{filename}", "wb") as f:
                        f.write(data)
                        logger.info(f"Successfully downloaded dataset to {str(dest)}/{filename}")

    @staticmethod
    def _download_file(url: str) -> bytes:
        response = requests.get(url)
        return response.content

    @staticmethod
    def _extract_filename(url: str) -> str:
        query = urllib.parse.urlsplit(url).query
        params = urllib.parse.parse_qs(query)
        return params["response-content-disposition"][0].split("=")[1].strip()

    def _generate_human_identities_data(self, request: DataRequest) -> DataResponse:
        params = dict(userId=self._user_id, orgId=self._org_id)
        response = requests.post(
            url=self._url + "/generations", params=params, headers=self._headers, json=request.dict()
        )
        if response.status_code != HTTPStatus.OK:
            message = f"Failed to generate a human identities data request with error: {json.loads(response.content)}"
            HttpStatusHandler.handle(status_code=response.status_code, message=message)

        logger.info("Successfully created a human identities data generation request.")
        response_info = json.loads(response.content)
        return DataResponse(
            generation_name=response_info["title"],
            renders=response_info["renders"],
            scenes=response_info["scenes"],
            generation_id=response_info["generationId"],
            dgu_hour=response_info["dguhCost"],
        )



from dataclasses import dataclass
from typing import TypeVar, Optional, Iterable

import marshmallow
import marshmallow_dataclass
import numpy as np
from marshmallow import pre_load
from marshmallow.fields import Field

SubSegments = TypeVar("SubSegments")


@dataclass
class _Segment:
    name: str


@dataclass
class NestedSegment(_Segment):
    sub_segments: Optional[SubSegments]

    def __getattr__(self, item):
        for sub_seg in self.sub_segments:
            if sub_seg.name == item:
                if isinstance(sub_seg, ColoredSegment):
                    return sub_seg.color
                else:
                    return sub_seg

    def __dir__(self) -> Iterable[str]:
        return [seg.name for seg in self.sub_segments]


@dataclass
class ColoredSegment(_Segment):
    color: Optional[np.ndarray] = None


class SubSegmentsField(Field):
    def _deserialize(self, value, *args, **kwargs):
        sub_segments = []
        for segment_name, segment_data in value.items():
            try:
                sub_segments.append(self._get_colored_segment(segment_name, segment_data))
            except KeyError:
                sub_segments.append(NestedSegment(segment_name, self._deserialize(segment_data, *args, **kwargs)))
        return sub_segments

    @staticmethod
    def _get_colored_segment(segment_name: str, segment_data: dict) -> ColoredSegment:
        if isinstance(segment_data, list):
            return ColoredSegment(segment_name, np.array(segment_data))
        else:
            return ColoredSegment(segment_name, np.array([segment_data["R"], segment_data["G"], segment_data["B"]]))


class SegmentSchema(marshmallow.Schema):
    TYPE_MAPPING = {SubSegments: SubSegmentsField}


@marshmallow_dataclass.dataclass(base_schema=SegmentSchema)
class Segmentation:
    scene: SubSegments

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs) -> dict:
        return {"scene": in_data}

    def __getattr__(self, item):
        for sub_seg in self.scene:
            if sub_seg.name == item:
                if isinstance(sub_seg, ColoredSegment):
                    return sub_seg.color
                else:
                    return sub_seg

    def __dir__(self) -> Iterable[str]:
        return [seg.name for seg in self.scene]


from dataclasses import field, dataclass
from typing import List, Iterator, Optional

from datagen import components
from datagen.components.datapoint import DataPoint
from datagen.components.datapoint import DatapointsContainer
from datagen.components import Scene
from datagen.components import DataSource
from datagen.components import SourcesRepository


@dataclass
class DatasetConfig:
    imaging_library: str = "opencv"
    environment: Optional[str] = None

    @property
    def override_environment(self) -> bool:
        return self.environment is not None


@dataclass
class Dataset:
    sources_repo: SourcesRepository = field(repr=False)
    config: DatasetConfig
    scenes: List[Scene] = field(default_factory=list, repr=False)

    def __post_init__(self):
        self._init_sources()

    def _init_sources(self) -> None:
        for source in self.sources_repo.get_all():
            self._init(source)

    def _init(self, source: DataSource) -> None:
        self._init_datapoints_container(source)
        self.scenes.extend(source.init_scenes())

    def _init_datapoints_container(self, source: DataSource) -> None:
        datapoints_container = DatapointsContainer(
            config={
                "environment": self.config.environment if self.config.override_environment else source.environment,
                "imaging_library": self.config.imaging_library
            }
        )
        datapoints_container.wire(packages=[components])

    def __iter__(self) -> Iterator[DataPoint]:
        for scene in self.scenes:
            for datapoint in scene:
                yield datapoint

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._get_single_datapoint(key)
        elif isinstance(key, slice):
            return self._get_multiple_datapoints(key)

    def _get_single_datapoint(self, idx: int) -> DataPoint:
        idx = idx % len(self)
        if not 0 <= idx < len(self):
            raise IndexError("Datapoint index exceeded dataset's size")
        [datapoint] = self._get_datapoints(start=idx, stop=idx + 1)
        return datapoint

    def _get_multiple_datapoints(self, key: slice) -> List[DataPoint]:
        start = key.start if key.start is not None else 0
        stop = key.stop % len(self) if key.stop is not None else len(self) - 1
        if not 0 <= start < stop < len(self):
            raise IndexError("Datapoints indexes exceeded dataset's size")
        return self._get_datapoints(start, stop)

    def _get_datapoints(self, start: int, stop: int) -> List[DataPoint]:
        datapoints = []
        for idx, datapoint in enumerate(self):
            if start <= idx < stop:
                datapoints.append(datapoint)
        return datapoints

    def __len__(self):
        return sum(len(scene) for scene in self.scenes)



import sys

SDK_PACKAGE_NAME = "datagen-tech"

if sys.version_info < (3, 10):
    from importlib_metadata import version as ver, PackageNotFoundError
else:
    from importlib.metadata import version as ver, PackageNotFoundError


def get_version():
    try:
        return ver(SDK_PACKAGE_NAME)
    except PackageNotFoundError:
        return None


sys.modules[__name__] = get_version()  # type: ignore



from pathlib import Path

from dynaconf import Dynaconf

SETTINGS_TOML_PATH = Path(__file__).parent.joinpath("settings.toml")

settings = Dynaconf(settings_files=[SETTINGS_TOML_PATH], environments=True)
